""" Non-Deterministic State Automaton (DFA)

A NFA is the tuple: <states, alphabet, tr, init, final>

WARNING: here we represent the alphabet indirectly with the 
set of atomic propositions AP, so alphabet = 2^AP


Example of use:


"""
from __future__ import annotations

import sys
from collections.abc import MutableMapping, Iterator, Iterable, Generator
from typing import Tuple, Optional, Collection, Set
from functools import reduce, partial

from mc308.logics.bool import BoolFormula
from mc308.auto.ts import TS, State

AutTrans = Tuple[State,Set[BoolFormula],State]


class NFA():
    def __init__(self : NFA,
                 states : Collection[State],
                 tr : Collection[AutTrans],
                 init : Collection[State],
                 final : Collection[State],
                 ap : Collection[BoolFormula]):
        """
        Creates a finite state automaton
        """

        self.states = set(states)
        self.init = set(init)
        assert(self.init <= self.states)
        self.final = set(final)
        assert(self.final <= self.states)

        self.ap = set(ap)
        assert reduce(lambda x,y : x and y.is_symbol(), self.ap, True)

        self.tr : MutableMapping[State, set[BoolFormula]]= {
            q : set() for q in self.states
        }
        for (s,a,p) in tr:
            assert s in self.states and p in self.states
            aset = frozenset(a)
            assert(aset <= self.ap)

            self.tr[s].add((aset,p))

    def __repr__(self : NFA) :
        pl = lambda l : ",".join(l)

        tr_list = set()
        for s in self.states:
            for (a,b) in self.tr.get(s, []):
                tr_list.add((s,a,b))

        return """NFA : {
  states : %s;
  init : %s;
  final : %s;
  tr : %s;
  ap : %s;
}""" % ( pl([str(l) for l in self.states]),
         pl([str(l) for l in self.init]),
         pl([str(l) for l in self.final]),
         pl(["(%s, %s, %s)" % (str(a[0]), str(a[1]), str(a[2])) for a in tr_list]),
         pl([str(a) for a in self.ap]),
        )


    def reverse(self : NFA) -> None :
        """
        Reverse the transitions of the NFA, invert inital and final.
        """
        tmp = self.init
        self.init = self.final
        self.final = tmp

        # reverse trans
        new_tr : MutableMapping[State, set[BoolFormula]]= {
            q : set() for q in self.states
        }
        for (s,dst) in self.tr.items():
            for (a,b) in dst:
                new_tr[b].add( (a,s) )

        self.tr = new_tr


    def prune_unreachable(self : NFA) -> None :
        """
        Prune the NFA removing all the unreachable states
        """
        stack = [i for i in self.init]
        visited : Set[State] = set()

        while len(stack) > 0:
            q = stack.pop()
            if not q in visited:
                visited.add(q)
                assert q in self.tr
                for (a,dst) in self.tr[q]:
                    stack.append(dst)


        self.states = self.states & visited
        self.final = self.final & visited
        new_tr : MutableMapping[State, set[BoolFormula]]= {
            q : set() for q in self.states
        }

        for (q, dest_set) in self.tr.items():
            if q in visited:
                dest_set_new = set()
                for (label, dest_state) in dest_set:
                    dest_set_new.add( (label, dest_state) )
                new_tr[q] = dest_set_new

        self.tr = new_tr

        if len(self.states) == 0:
            self.states = set([1])
            self.tr[1] = set()


    def prune(self : NFA) -> None :
        """
        Prune the NFA removing all unuseful states
        """
        self.reverse()
        self.prune_unreachable()
        self.reverse()
        self.prune_unreachable()

    def is_empty(self : NFA) -> bool :
        self.prune()
        return len(self.final) == 0

